from inspect import Parameter, Signature, getmembers, iscoroutinefunction, ismethod, signature
from typing import Any, Callable, Dict, List, Tuple

from fastapi import APIRouter

from .utils import is_overridden_func


class ManageSignature:
    """A class for interacting with a function signature."""

    _required_args: List[str]

    @staticmethod
    def __get_route_kwargs__(func: Callable) -> Dict[str, Any]:
        """
        Get default values of route_kwargs from func.

        Parameters
        ----------
        func: Target func

        Returns
        -------
        Kwargs
        """
        sign = signature(func)
        res = next(filter(lambda x: x.name == "route_kwargs", sign.parameters.values()), {})
        if res:
            return res.default
        return res

    def __replace_signature__(self, func: Callable) -> Callable:
        """
        Replace signature of func.

        Parameters
        ----------
        func: Target func

        Returns
        -------
        New function with new signature.
        """

        def route_handler(*args, **kwargs):
            return func(*args, **kwargs)

        async def async_route_handler(*args, **kwargs):
            return await func(*args, **kwargs)

        new_func = async_route_handler if iscoroutinefunction(func) else route_handler
        sign = signature(func)
        new_params = self.__gen_new_params__(sign)
        new_func.__signature__ = sign.replace(parameters=new_params)
        return new_func

    @staticmethod
    def __gen_new_params__(sign: Signature) -> List[Parameter]:
        """
        Generate list with new params for signature.

        Parameters
        ----------
        sign: Signature

        Returns
        -------
        List with new params.
        """
        old_params = list(sign.parameters.values())
        return [p.replace(kind=Parameter.POSITIONAL_OR_KEYWORD) for p in old_params if p.name != "route_kwargs"]


class HTTPMethods:
    """The supported HTTP methods are described here."""

    _HTTP_METHODS: tuple = ("head", "options", "get", "post", "patch", "put", "delete")

    def _get_route_handlers(self) -> List[Tuple[str, Callable]]:
        """
        Get route handler functions.

        Returns
        -------
        List with data o route handlers.
        """
        filter_obj = filter(
            lambda x: x[0] in self._HTTP_METHODS and is_overridden_func(x[1]), getmembers(self, predicate=ismethod)
        )
        return list(filter_obj)

    def head(self, **kwargs):
        ...

    def options(self, **kwargs):
        ...

    def get(self, **kwargs):
        ...

    def post(self, **kwargs):
        ...

    def patch(self, **kwargs):
        ...

    def put(self, **kwargs):
        ...

    def delete(self, **kwargs):
        ...


class Resource(HTTPMethods, ManageSignature):
    """This class is designed to implement CRUD methods of a specific api resource with FastAPI."""

    router: APIRouter
    tag: str = None
    path: str = ""

    def __init__(self, path: str = None):
        """
        Init Resource instance.
        If the value of the 'tag' attribute is not set. Then the value will be the name of the class instance.

        Parameters
        ----------
        path: Path for this resource
        """
        if self.tag is None:
            self.tag = self.__class__.__name__
        if path is not None:
            self.path = path
        self.__init_router__()

    def __init_router__(self) -> None:
        """
        Init APIRouter of FastAPI and include implemented methods.

        Returns
        -------
        None
        """
        self.router = APIRouter(prefix=self.path, tags=[self.tag])
        self.__include_methods__()

    def __include_methods__(self):
        """
        Include HTTP methods to router.

        Get kwargs from implemented method (only for route), and replace signature of route_handler,
        then add api route to router.

        Returns
        -------
        None
        """
        for method, handler in self._get_route_handlers():
            kwargs = {"summary": method}
            kwargs.update(self.__get_route_kwargs__(handler))
            route_handler = self.__replace_signature__(handler)
            self.router.add_api_route(path="", endpoint=route_handler, methods=[method.capitalize()], **kwargs)

    @property
    def _required_args(self) -> List[str]:
        """
        Get required arg names from path of this Resource instance.

        Returns
        -------
        List with required arg names.
        """
        return [item.strip("{}") for item in self.path.split("/") if item.startswith("{") and item.endswith("}")]



from inspect import Parameter, signature
from typing import Callable


def is_overridden_func(target_func) -> bool:
    """
    Check if the function is overridden in child class.

    Parameters
    ----------
    target_func: Target function

    Returns
    -------
    True/False
    """
    child_instance = target_func.__self__
    func_of_parent = getattr(super(type(child_instance), child_instance), target_func.__name__)
    return target_func.__func__ != func_of_parent.__func__


def route_settings(**kwargs) -> Callable:
    """
    Decorator for forwarding 'kwargs' to the 'add_api_route' method of the 'APIRouter` instance.

    Parameters
    ----------
    kwargs: Kwargs for add_api_route.
            See descriptions for all kwargs in `fastapi.APIRouter.add_api_route`.

    Returns
    -------
    Func with new signature.
    """

    def decorator(func):
        route_kwargs_param = Parameter(name="route_kwargs", kind=Parameter.POSITIONAL_OR_KEYWORD, default=kwargs)
        sign = signature(func)
        old_params = list(sign.parameters.values())
        old_params.append(route_kwargs_param)
        func.__signature__ = sign.replace(parameters=old_params)
        return func

    return decorator



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastapi_restful']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.78.0', 'pre-commit>=2.20.0,<3.0.0']

setup_kwargs = {
    'name': 'fastapi-restful-extension',
    'version': '0.4.1',
    'description': 'Extension for make RESTful interfaces with FastAPI.',
    'long_description': '<p align="center">\n    <a href="https://github.com/maximshumilo/fastapi-restful-extension/actions?query=workflow%3ATest+event%3Apush+branch%3Amaster" target="_blank">\n        <img src="https://github.com/maximshumilo/fastapi-restful-extension/actions/workflows/test.yml/badge.svg">\n    </a>\n    <a href="https://codecov.io/gh/maximshumilo/fastapi-restful-extension">\n        <img src="https://img.shields.io/codecov/c/gh/maximshumilo/fastapi-restful-extension?color=31c955"/>\n    </a>\n    <a href="https://pypi.org/project/fastapi-restful-extension/" target="_blank">\n        <img src="https://img.shields.io/pypi/v/fastapi-restful-extension?color=31c955&label=pypi%20package">\n    </a>\n    <img alt="PyPI - Python Version" src="https://img.shields.io/pypi/pyversions/fastapi-restful-extension?color=31c955">\n</p>\n\n<p align="center">\n    FastAPI-RESTful-Extension - is extension for FastAPI that allows you to easily create a REST API.\n</p>\n\n---\n\n**Documentation**: <a href="https://maximshumilo.github.io/fastapi-restful-extension/" target="_blank">https://maximshumilo.github.io/fastapi-restful-extension/ </a>\n\n**PyPI**: <a href="https://pypi.org/project/fastapi-restful-extension/" target="_blank">https://pypi.org/project/fastapi-restful-extension/ </a>\n\n---\n\n# Installation\n\nFastAPI-RESTful-Extension has the following dependencies:\n\n- FastAPI <= `0.78.0`\n- Python version: `3.7, 3.8, 3.9` or `3.10`\n\n---\n\n### From PyPI\nInstall package with pip from `pypi.org`\n\n```console\n$ pip install fastapi-restful-extension\n\nCollecting fastapi-restful-extension\n  Downloading fastapi_restful_extension-0.3.0-py3-none-any.whl (8.0 kB)\n...\n...\n...\nInstalling collected packages: fastapi-restful-extension\nSuccessfully installed fastapi-restful-extension-0.3.0\n```\n\n---\n',
    'author': 'Shumilo Maksim',
    'author_email': 'shumilo.mk@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Dict, Optional, Type, Union

from fastapi import APIRouter, FastAPI

from .resource import Resource


class APIMixin:
    router: APIRouter

    @staticmethod
    def _strip_prefix(prefix: Optional[str]) -> str:
        return prefix.strip("/") if prefix is not None else ""

    @property
    def urls(self) -> dict:
        """
        Registered urls and HTTP methods.

        Returns
        -------
        Url map
        """
        urls = dict()
        for route in self.router.routes:
            urls.setdefault(route.path, set())
            urls[route.path].update(route.methods)
        return urls

    def add_resource(self, resource: Type[Resource], path: str = None) -> None:
        """
        Initial resource and add resource router to FastAPI.

        Parameters
        ----------
        resource
            Type of Resource
        path
            Resource path

        Returns
        -------
        None
        """
        resource_instance = resource(path)
        self.router.include_router(resource_instance.router)


class APIVersion(APIMixin):
    """
    A class specific version API.
    """

    def __init__(self, version_prefix: str):
        prefix = self._strip_prefix(version_prefix)
        self.router = APIRouter(prefix=f"/{prefix}")

    def __str__(self) -> str:
        return self.router.prefix


class RestAPI(APIMixin):
    """Main class for create RESTful-API."""

    router: Union[APIRouter, FastAPI]

    def __init__(self, fastapi_app: FastAPI, prefix: Optional[str] = "/api"):
        """
        Init class.

        Parameters
        ----------
        fastapi_app: instance of FastAPI
        prefix: Default prefix in url path.
        """
        self._fastapi: FastAPI = fastapi_app
        self._prefix: str = prefix
        self._versions: Dict[str, APIVersion] = {}
        self.__init_main_router__()

    def __getitem__(self, item: str) -> Optional[APIVersion]:
        """
        Get instance API by prefix version.

        Parameters
        ----------
        item: Prefix version

        Returns
        -------
        Instance API or None
        """
        prefix = self._strip_prefix(item)
        return self._versions.get(f"/{prefix}")

    def __init_main_router__(self) -> None:
        """
        Init main router.

        Returns
        -------
        None
        """
        prefix = self._strip_prefix(self._prefix)
        if prefix:
            router = APIRouter(prefix=f"/{prefix}")
        else:
            router = self._fastapi
        self.router = router

    @property
    def versions(self) -> Dict[str, APIVersion]:
        """Versions map"""
        return self._versions

    def apply(self) -> None:
        """
        Include new api urls to FastAPI app.

        Returns
        -------
        None
        """
        for router in self._versions.values():
            self.router.include_router(router.router)
        if isinstance(self.router, APIRouter):
            self._fastapi.include_router(self.router)

    def include_api_version(self, api_version: APIVersion) -> None:
        """
        Include API version to main router.

        Parameters
        ----------
        api_version
            Instance of APIVersion with included resources.
        """
        prefix = api_version.router.prefix
        if prefix in self._versions:
            raise AssertionError(f"This version is exist: {prefix}")
        self._versions[prefix] = api_version
        self.router.include_router(api_version.router)
